<?php

namespace Teknisa\EST\Services;
use Zeedhi\Framework\DataSource\FilterCriteria;
use Zeedhi\Framework\DTO;
use Teknisa\Model\Estoque\Entities\Lanctoestoq;
use Teknisa\Libs\Util\NewCode;
use Teknisa\Libs\Util\DateUtil;
use Zeedhi\Framework\DTO\Row;
use Zeedhi\Framework\DataSource\DataSet;
use Zeedhi\Framework\DTO\Response;

class ProdProcessaService {
    
    protected $parameterBagColumns = array("CDFILIAL", "NRLANCPROCESSA", "CDPRODUTO");
    protected $entityManager;
    protected $crudService;
    protected $environment;
    protected $connection;
    protected $newCode;
    public $arrProdIrregulares = array();

    public function __construct($entityManager, $crudService, $environment, $newCode, $validacoesEstoqueService) {
        $this->entityManager = $entityManager;
        $this->crudService = $crudService;
        $this->environment = $environment;
        $this->connection = $entityManager->getConnection();
        $this->newCode = $newCode;
        $this->validacoesEstoqueService = $validacoesEstoqueService;
    }

    public function doStockOperations(&$retirada) {
        try{
            $boolError;
            $connection = $this->entityManager->getConnection();
            $connection->beginTransaction();  //Desativa auto commit
            
            //Parte Referente a Entrada.
            //realiza o lançamento de estoque  referente a saída e entrada  ------ 4-Saida 2 -Entrada
            $nrctrlmovest = $retirada['NRLANCPROCESSA'];
            $idtplanctoOut = $retirada['CDTIPORETI'] ? 3 : 4;
            $nrlancamentsaida    = $this->saveTabelaLancamentoEstoque($retirada,$nrctrlmovest,$idtplanctoOut);
            $nrlancamentoentrada = $this->saveTabelaLancamentoEstoque($retirada,$nrctrlmovest,2);
            $retirada['NRLANCESTQSAID'] = $nrlancamentsaida;
            $retirada['NRLANCESTQENTR'] = $nrlancamentoentrada;
            $this->updateLanctoProcessaWithFkNumbers($retirada);
            
            // realiza retirada na tabela retiradaesto no caso da operação ser uma retirada
            if($idtplanctoOut == 3){
                $this->saveTabelaRetiradaEstoque($retirada);
            }
            
            $conditions =  array();
            $filterRow = new FilterCriteria('itlancprocessa_produto', $conditions);
            $filterRow->setDataSourceName('itlancprocessa_produto');
            $filterRow->addCondition('CDFILIAL', $retirada['CDFILIAL']);
            $filterRow->addCondition('NRLANCPROCESSA', $retirada['NRLANCPROCESSA']);
            $filterRow->addCondition('CDPRODUTO', $retirada['CDPRODUTO']);
            $newFilterCriteria = $this->crudService->getPopulatedFilterCriteria($filterRow, $this->parameterBagColumns);
            $rowResponse = new DTO\Response;
            $this->crudService->find($rowResponse, $newFilterCriteria);
            $rowsEntrada = $rowResponse->getDataSets()[0]->getRows();
            
            //realiza o lançamento de item de saída e entrada  ------ 4-Saida 2 -Entrada
            $this->saveTabelaItemLancamentoEstoque($nrctrlmovest, $retirada, [$retirada], $idtplanctoOut, $nrlancamentsaida);
            $this->saveTabelaItemLancamentoEstoque($nrctrlmovest, $retirada, $rowsEntrada, 2, $nrlancamentoentrada);
            
            $this->atualizaEstoque($retirada['CDFILIAL'], $nrlancamentsaida, $nrlancamentoentrada);
            
            $connection->commit();
            return $retirada;
        }catch (\Exception $e) {
          $boolError = $e;
          $this->entityManager->rollback(); //rollback em caso de erro
          $irregulares = $this->validacoesEstoqueService->getIrregularProducts();
          foreach($irregulares as $irregular){
            if($retirada['CDPRODESTO'] == $irregular->getCdproduto()){
                $retirada['error'] = $e;
                $retirada['errorMessage'] = $e->getMessage();
                array_push($this->arrProdIrregulares, $retirada);
            }
          }
       }
      
      if(count($this->arrProdIrregulares) > 0){
          throw $this->arrProdIrregulares[0]['error'];
      }
      if(!empty($boolError))
        throw $boolError;
      /*$dtProcessa = new \DateTime($fullDate[2].'-'.$fullDate[1].'-'.$fullDate[0]);
      $paramfilial = $this->entityManager->getRepository('\Geral\Entities\Paramfilial')
        ->findOneBy(
          array(
            'cdfilial' => $cdfilial
          )
        );
      $paramfilial->setDtprocessa($dtProcessa);
      $this->entityManager->persist($paramfilial);
      $this->entityManager->flush();*/
      
    }
    
    protected function updateLanctoProcessaWithFkNumbers($retirada) {
        $retirada['__is_new'] = false;
        $row = new Row();
        foreach($retirada as $key => $value){
            $row->set($key, $value);
        }
        $this->crudService->save('lancprocessa', [$row]);
        $this->entityManager->flush();
    }
    
    protected function atualizaEstoque($cdfilial, $nrlanestsaid, $nrlanestentr) {
        $nrorg = $this->environment->getNrOrgTrab();
        $lancsaida = $this->validacoesEstoqueService->buscaLanctoestoq($cdfilial, $nrlanestsaid, $nrorg);
        $this->validacoesEstoqueService->setLanctoestoq($lancsaida);
        $this->validacoesEstoqueService->atualizaEstoque($cdfilial, $nrlanestsaid);
        $this->validacoesEstoqueService->setLanctoestoq(\Estoque\Entities\Lanctoestoq::findOneByCdfilialAndNrlancestqAndNrorg($cdfilial, $nrlanestentr, $nrorg));
        $this->validacoesEstoqueService->atualizaEstoque($cdfilial, $nrlanestentr);
    }
    
    public function saveTabelaLancamentoEstoque($row,$nrctrlmovest,$idtplancto){
      
        $nrorg = $this->environment->getNrOrgTrab();
        $operador = $this->environment->getCdOperador();
        $currentTime = $this->environment->getCurrentTime();
        
        $nrlancestq = $this->newCode->getCodeByTableName('LANCTOEST'.$row['CDFILIAL'], $nrorg ,1,10);
        
        $lanctoestoq = new Lanctoestoq();
        //campos auditoria
        $lanctoestoq->setNrorg($nrorg);
        $lanctoestoq->setDtinclusao($currentTime);
        $lanctoestoq->setCdoperinclusao($operador);
        $lanctoestoq->setNrorginclusao($nrorg);
        $lanctoestoq->setNrorgultatu($nrorg);
        $lanctoestoq->setDtultatu($currentTime);
        $lanctoestoq->setCdoperultatu($operador);
        $lanctoestoq->setIdativo("S");
        
        $lanctoestoq->setCdfilial($row['CDFILIAL']);
        $lanctoestoq->setNrlancestq($nrlancestq);
        $lanctoestoq->setIdtplancto($idtplancto);
        $data = substr($row['DTLANCPROCESSA'], 0, 10);
        $lanctoestoq->setDtlancestq(DateUtil::getDataDeString($data));
        //Processamento 
        if($idtplancto == 3){
            $lanctoestoq->setDslancestq($row['DSLANCESTQ'], 0, 24);
        }else{
            $lanctoestoq->setDslancestq('Processamento '.$nrctrlmovest.' do produto '.substr($row['NMPRODUTO'], 0, 24));
        }
        $lanctoestoq->setCdfilimovi($row['CDFILIAL']);
        $lanctoestoq->setCdoperador($operador);
        
        $this->entityManager->persist($lanctoestoq);
        $this->entityManager->flush();
        
        return $nrlancestq;
    }
    
    public function saveTabelaItemLancamentoEstoque($nrctrlmovest, $rowEntrada, $produtos, $identradaSaida, $nrlancamentoestoque){
        try {
            $nrorg       = $this->environment->getNrorgTrab();
            $operador    = $this->environment->getCdOperador();
            $currentTime = $this->environment->getCurrentTime()->format("d/m/Y H:i:s");
            $arrayItens  = [];
            $paramfilial = $this->buscaTipoCustoFilial($nrorg,$rowEntrada['CDFILIAL']);
            $countproduto = 0;
            
            foreach ($produtos as &$prod) {
                $cdalmoxarife = '';
                $cdlocalestoq = '';
                $paramProduto = $this->buscaParamsProdfili($nrorg, $rowEntrada['CDFILIAL'], $prod['CDPRODESTO']);

                //item de saida ou retirada
                if($identradaSaida == 4 || $identradaSaida == 3 ){
                    if($paramfilial[0]['IDUTILALMOX'] == 'S')
                        $cdalmoxarife = $paramProduto[0]['CDMOXPRODSERPROC']?? ($paramfilial[0]['CDALMOXARIFE']?? ' ');
                    else
                        $cdalmoxarife = ' ';
                    if($paramfilial[0]['IDUTILLCESTQ'] == 'S')
                        $cdlocalestoq = $paramProduto[0]['CDLOCPRODSERPROC']?? ($paramfilial[0]['CDLOCALESTOQ']?? ' ');
                    else
                        $cdlocalestoq = ' ';
                        
                    $prod['QTTOTLANCTO'] = $prod['QTLANCPROCESSA'];
                }
                //item de entrada
                else{
                    if($paramfilial[0]['IDUTILALMOX'] == 'S')
                        $cdalmoxarife = $paramProduto[0]['CDALMOXPRODPROC']?? ' ';
                    else
                        $cdalmoxarife = ' ';
                    if($paramfilial[0]['IDUTILLCESTQ'] == 'S')
                        $cdlocalestoq = $paramProduto[0]['CDLOCALPRODPROC']?? ' ';
                    else
                        $cdlocalestoq = ' ';
                        
                    $prod['QTTOTLANCTO'] = $prod['QTITPROCESSA'];
                }
                
                if($prod['QTTOTLANCTO'] > 0){
                    $prod['NRLOTEESTQ'] = $prod['NRLOTEESTQ']?? ' ';
    
                    $prod['NRSUBLOTE'] = $prod['NRSUBLOTE']?? ' ';
    
                    $prod['__is_new'] = false;
    
                    $cdalmoxarifeCusto = $rowEntrada['CDALMOXSAID']?? ' ';
                    $cdlocalestoqCusto = $rowEntrada['CDLCESTSAID']?? ' ';
                    
                    $vlrunitario  =  $prod['VRUNITLANC']?? $prod['VRUNITITPROCESSA'];
    
                    $countproduto++;
                    $nrsequitem   = str_pad($countproduto, 4, "0", STR_PAD_LEFT);
                    
                    $row = new Row();
                    $row->set('__is_new',     true);
                    $row->set('NRORG',        $nrorg);
                    $row->set('NRORGULTATU',  $nrorg);
                    $row->set('NRORGINCLUSAO', $nrorg);
                    $row->set('DTULTATU',     $currentTime);
                    $row->set('DTINCLUSAO',   $currentTime);
                    $row->set('CDOPERINCLUSAO', $operador);
                    $row->set('CDOPERULTATU', $operador);
                    $row->set('IDATIVO',      'S');
                    $row->set('CDOPERADOR',   $operador);
                    $row->set('CDFILIAL',     $rowEntrada['CDFILIAL'] );
                    $row->set('NRLANCESTQ',   $nrlancamentoestoque);
                    $row->set('NRSEQUITEM',   $nrsequitem);
                    $row->set('CDPRODUTO',    $prod['CDPRODUTO']);
                    $row->set('CDALMOXARIFE', $cdalmoxarife);
                    $row->set('CDLOCALESTOQ', $cdlocalestoq);
                    
                    if($identradaSaida == 3)
                        $row->set('DSLANCESTIT',  $rowEntrada['DSLANCESTQ']);
                    else
                        $row->set('DSLANCESTIT',  'Processamento '.$nrctrlmovest.' do produto '.substr($rowEntrada['NMPRODUTO'], 0, 24));
                    
                    $row->set('NRLOTEESTQ',   $prod['NRLOTEESTQ']);
                    $row->set('NRSUBLOTE',    $prod['NRSUBLOTE']);
                    $row->set('CDFILIMOVI',   $rowEntrada['CDFILIAL']);
                    $row->set('CDPRODMOVI',   $prod['CDPRODESTO']);
                    $row->set('IDORDEMCLC',   $identradaSaida);
                    $row->set('DTLANCMOVI' ,  substr($rowEntrada['DTLANCPROCESSA'], 0, 10));
                    $row->set('IDEVOLESTOQ',  'S');
                    $row->set('IDTIPOMOVI',   $identradaSaida);
                    $row->set('QTTOTLANCTO',  floatval ($prod['QTTOTLANCTO']));
                    $row->set('QTLANCTOEST',  floatval ($prod['QTTOTLANCTO']) * ($prod['VRFATOCONV'] == 0 ? 1 : $prod['VRFATOCONV']));
                    $row->set('VRUNILANCTO' , $vlrunitario);
                    $row->set('VRTOTLANCTO' , $vlrunitario * floatval($prod['QTTOTLANCTO']));
                    $row->set('VRESTOQDIA' ,  $vlrunitario * floatval($prod['QTTOTLANCTO']));
                    $row->set('VRESTOQBRUT' , ($vlrunitario / ($prod['VRFATOCONV'] == 0 ? 1 : $prod['VRFATOCONV'])) * floatval($prod['QTTOTLANCTO']));
                    $row->set('VRLANCTOEST' , $vlrunitario / ($prod['VRFATOCONV'] == 0 ? 1 : $prod['VRFATOCONV']));
                    $row->set('VRLANCTOBRUT', $vlrunitario / ($prod['VRFATOCONV'] == 0 ? 1 : $prod['VRFATOCONV']));
                    $row->set('NRLANCORIGRP',  null);
                    $row->set('NRSEQITEMREQEST',  null);
                    $row->set('QTDEVOLESTOQ',  null);
                    $arrayItens [] = $row;
                }
            }
            $this->crudService->save('itlanctoest_produto_itrequesto', $arrayItens);
            $this->entityManager->flush();
        }catch (\Exception $e) {
            throw $e;
        }
    }
    
    protected function saveTabelaRetiradaEstoque($retirada) {
        $nrorg = $this->environment->getNrOrgTrab();
        $operador = $this->environment->getCdOperador();
        
        $cdcontador = "RETIESTO".$retirada["CDFILIAL"];
        $retirada["NRRETEST"]  = $this->newCode->getCodeByTableName($cdcontador, $nrorg, 1, 10);

        $row = new Row();
        $row->set("CDFILIAL",$retirada["CDFILIAL"]);
        $row->set("NRRETEST",$retirada["NRRETEST"]);
        $row->set("DTRETEST",$retirada["DTLANCPROCESSA"]);
        $row->set("DSRETEST",$retirada["DSLANCESTQ"]);
        $row->set("CDFILIALLAN",$retirada["CDFILIAL"]);
        $row->set("NRLANCESTQ",$retirada["NRLANCESTQSAID"]);
        $row->set("CDSETOR",$retirada["CDSETOR"]);
        $row->set("CDCENTCUST",$retirada["CDCENTCUST"]);
        $row->set('NRORDPROD', null);
        $row->set('NRSUBORDPROD', null);
        $row->set('CDETAPAPROD', null);
        $row->set('NRREQUESTO', null);
        $row->set('IDPESPRODRET', 'N');
        $row->set('CDFILIALORE', null);
        $row->set('NRORCAEVENTO', null);
        $row->set('NRSEQORCAEVE', null);
        $row->set('IDRETIRAPERDA', 'N');
        $row->set('IDAGUARETIRADA', null);
        $row->set('NRORG', $nrorg);
        $row->set('IDATIVO', 'S');
        $row->set('DTINCLUSAO', DateUtil::getDataAtual()->format('d/m/Y'));
        $row->set('DTULTATU', DateUtil::getDataAtual()->format('d/m/Y'));
        $row->set('NRORGINCLUSAO', $nrorg);
        $row->set('CDOPERINCLUSAO', $operador);
        $row->set('NRORGULTATU', $nrorg);
        $row->set('CDOPERULTATU', $operador);
        $row->set("__is_new",true);

        $this->crudService->save('retiradaesto', [$row]);
        $this->entityManager->flush();
    }
    
    public function buscaTipoCustoFilial($nrorg, $cdfilial) {
        $sql = "SELECT NVL(CDTIPCUSTEST,3) AS CDTIPCUSTEST, CDALMOXARIFE, CDLOCALESTOQ,
                IDUTILALMOX, IDUTILLCESTQ, IDUTILLOTE
                FROM PARAMFILIAL
                WHERE
                    CDFILIAL = :P_CDFILIAL AND
                    NRORG    = :P_NRORG";
        $params = array(
           'P_NRORG'=>$nrorg,
           'P_CDFILIAL'=>$cdfilial
        );

        $std = $this->connection->prepare($sql);
        $std->execute($params);
        return $std->fetchAll();
    }
    
    public function buscaParamsProdfili($nrorg, $cdfilial, $cdproduto) {
        $sql =  "SELECT CDMOXPRODSERPROC,
                        CDLOCPRODSERPROC,
                        CDALMOXPRODPROC,
                        CDLOCALPRODPROC
                FROM PRODFILI
                WHERE
                    CDPRODUTO = :P_CDPRODUTO AND
                    CDFILIAL = :P_CDFILIAL AND
                    NRORG    = :P_NRORG";
        $params = array(
           'P_NRORG'=>$nrorg,
           'P_CDFILIAL'=>$cdfilial,
           'P_CDPRODUTO'=>$cdproduto
        );

        $std = $this->connection->prepare($sql);
        $std->execute($params);
        return $std->fetchAll();
    }
    
    public function getIrregularProducts() {
        return $this->arrProdIrregulares;
    }
    
    public function deleteStockOperations($retirada) {
        try{
            $boolError;
            $nrorg = $this->environment->getNrOrgTrab();
            $connection = $this->entityManager->getConnection();
            $connection->beginTransaction();  //Desativa auto commit

            $lanctoRetirada = $this->validacoesEstoqueService->buscaLanctoestoq($retirada["CDFILIAL"], $retirada["NRLANCESTQSAID"], $nrorg);
            $this->deletaRetiradaEsto($retirada["CDFILIAL"],$lanctoRetirada,$nrorg,$retirada);
            
            $lanctoEntrada = $this->validacoesEstoqueService->buscaLanctoestoq($retirada["CDFILIAL"], $retirada["NRLANCESTQENTR"], $nrorg);
            $this->validacoesEstoqueService->deletaLancamentoEstoq($lanctoEntrada);
            $this->validacoesEstoqueService->atualizaEstoque($retirada["CDFILIAL"], $retirada["NRLANCESTQENTR"]);
            
            $lanctoSaida = $this->validacoesEstoqueService->buscaLanctoestoq($retirada["CDFILIAL"], $retirada["NRLANCESTQSAID"], $nrorg);
            $this->validacoesEstoqueService->deletaLancamentoEstoq($lanctoSaida);
            $this->validacoesEstoqueService->atualizaEstoque($retirada["CDFILIAL"], $retirada["NRLANCESTQSAID"]);
            
            $retirada['NRLANCESTQENTR'] = null;
            $retirada['NRLANCESTQSAID'] = null;
            $this->updateLanctoProcessaWithFkNumbers($retirada);
            
            $this->entityManager->flush();
            $connection->commit();
        } catch(\Exception $e){
            $boolError = $e;
            $this->entityManager->rollback(); //rollback em caso de erro
            $irregulares = $this->validacoesEstoqueService->getIrregularProducts();
            foreach($irregulares as $irregular){
                if($retirada['CDPRODESTO'] == $irregular->getCdproduto()){
                    $retirada['error'] = $e;
                    $retirada['errorMessage'] = $e->getMessage();
                    array_push($this->arrProdIrregulares, $retirada);
                }
            }
        }
        
        if(count($this->arrProdIrregulares) > 0){
            throw $this->arrProdIrregulares[0]['error'];
        }
        if(!empty($boolError))
            throw $boolError;
    }
    
    public function deletaRetiradaEsto($cdfilial,$lanctoRetirada,$nrorg,$retirada){
        
        $filterRow = new FilterCriteria('retiradaesto', []);
        $filterRow->addCondition('CDFILIAL', $cdfilial);
        $filterRow->addCondition('NRLANCESTQ', $lanctoRetirada->getNrlancestq());
        $filterRow->addCondition('NRORG', $nrorg);
        $newFilterCriteria = $this->crudService->getPopulatedFilterCriteria($filterRow, $this->parameterBagColumns);
        $rowResponse = new DTO\Response;
        $this->crudService->find($rowResponse, $newFilterCriteria);
        $nrretest = $rowResponse->getDataSets()[0]->getRows()[0]['NRRETEST'];
        
        if($lanctoRetirada->getIdtplancto() == 3){
            $row = [
                'CDFILIAL' => $cdfilial,
                'NRLANCESTQ' => $lanctoRetirada->getNrlancestq(),
                'NRORG' => $nrorg,
                'IDRETIRAPERDA' => 'N',
                'DTRETEST' => $retirada['DTLANCPROCESSA'],
                'NRRETEST'=> $nrretest,
                'IDATIVO' => 'S',
                '__is_new' => false
            ];
            
            $dataSet = new DataSet('retiradaesto', [$row]);
            $response = new DTO\Response;
            $this->crudService->delete($response, $dataSet);
        }
    }
}